﻿using EFInfrastructure.Helpers;
using Microsoft.EntityFrameworkCore;
using System.Text;

namespace EFInfrastructure
{
    public class EFQuery<TEntity, TKey> where TEntity : class, IEntity<TKey>, new()
    {
        private const int defaultPageSize = 10;
        private string orderBy;
        private DbContext context;

        public int PageSize { get; private set; } = defaultPageSize;

        public int DesiredPage { get; private set; } = 0;

        public string OrderBy
        {
            get => orderBy;
            private set
            {
                orderBy = GetPropertyNameIfExists(value);
            }
        }

        public bool UseAscendingOrder { get; private set; } = true;

        public MyPredicate Predicate { get; private set; }

        public EFQuery(DbContext context)
        {
            this.context = context;
        }

        public EFQuery<TEntity, TKey> Where(MyPredicate predicate)
        {
            Predicate = predicate ?? throw new ArgumentException("Predicate must be defined!");
            return this;
        }

        public EFQuery<TEntity, TKey> SortBy(string orderBy, bool ascendingOrder = true)
        {
            OrderBy = !string.IsNullOrWhiteSpace(orderBy) ? orderBy : throw new ArgumentException($"{nameof(orderBy)} must be defined!");
            UseAscendingOrder = ascendingOrder;
            return this;
        }

        public EFQuery<TEntity, TKey> Page(int pageToFetch, int pageSize = defaultPageSize)
        {
            if (pageToFetch < 1)
            {
                throw new ArgumentException("Desired page number must be greater than zero!");
            }
            if (pageSize < 1)
            {
                throw new ArgumentException("Page size must be greater than zero!");
            }
            DesiredPage = pageToFetch;
            PageSize = pageSize;
            return this;
        }

        public EFQueryResult<TEntity, TKey> Execute()
        {
            EFQueryResult<TEntity, TKey> result;
            var sql = new StringBuilder()
                .Append($"{SqlConstants.SelectFromClause}dbo.{typeof(TEntity).Name}s");

            if (Predicate != null)
            {
                var predicateResult = Predicate.BuildSimplePredicate();

                sql.Append($"{SqlConstants.WhereClause}{predicateResult}");
            }

            var itemsCount = context.Set<TEntity>().FromSqlRaw(sql.ToString()).Count();

            if (!string.IsNullOrWhiteSpace(OrderBy))
            {
                sql.Append(SqlConstants.OrderByClause + OrderBy + (UseAscendingOrder ? SqlConstants.Ascending : SqlConstants.Descending));
            }

            if (DesiredPage > 0)
            {
                var items = context.Set<TEntity>().FromSqlRaw(sql.ToString()).ToList().Skip((DesiredPage - 1) * PageSize).Take(PageSize).ToList();
                result = new EFQueryResult<TEntity, TKey>(items, itemsCount, PageSize, DesiredPage);
            }
            else
            {
                var items = context.Set<TEntity>().FromSqlRaw(sql.ToString()).ToList();
                result = new EFQueryResult<TEntity, TKey>(items, itemsCount);
            }
            return result;
        }

        private static string GetPropertyNameIfExists(string input)
        {
            var properties = typeof(TEntity).GetProperties().Select(prop => prop.Name);
            var matchedName = properties
                .FirstOrDefault(name => name.IndexOf(input, StringComparison.OrdinalIgnoreCase) >= 0);
            return matchedName ?? string.Empty;
        }
    }
}
